#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using System.Collections.Generic;
using HeuristicLab.Problems.DataAnalysis.Symbolic.Symbols;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.Symbols;

namespace HeuristicLab.Problems.DataAnalysis.Regression.Symbolic.Analyzers {
  /// <summary>
  /// An operator that analyzes the population diversity using fine grained structural tree similarity estimation.
  /// </summary>
  [Item("FineGrainedStructuralPopulationDiversityAnalyzer", "An operator that analyzes the population diversity using fine grained structural tree similarity estimation.")]
  [StorableClass]
  public sealed class FineGrainedStructuralPopulationDiversityAnalyzer : SymbolicRegressionPopulationDiversityAnalyzer {

    [StorableConstructor]
    private FineGrainedStructuralPopulationDiversityAnalyzer(bool deserializing) : base(deserializing) { }
    private FineGrainedStructuralPopulationDiversityAnalyzer(FineGrainedStructuralPopulationDiversityAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public FineGrainedStructuralPopulationDiversityAnalyzer() : base() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new FineGrainedStructuralPopulationDiversityAnalyzer(this, cloner);
    }

    protected override double[,] CalculateSimilarities(SymbolicExpressionTree[] solutions) {
      int n = solutions.Length;
      List<string> variableNames = new List<string>();
      foreach (StringValue variableName in ProblemData.InputVariables) {
        variableNames.Add(variableName.Value);
      }
      variableNames.Add(ProblemData.TargetVariable.Value);
      IList<GeneticInformationItem>[] geneticInformationItemsList = new List<GeneticInformationItem>[n];
      for (int i = 0; i < n; i++) {
        geneticInformationItemsList[i] = GeneticInformationItem.getGeneticInformationItems(solutions[i].Root, variableNames, 1, 5);
      }
      double[,] result = new double[n, n];
      for (int i = 0; i < n; i++) {
        for (int j = 0; j < n; j++) {
          if (i == j)
            result[i, j] = 1;
          else
            result[i, j] = 0;
        }
      }
      return result;
    }


    #region private class GeneticInformationItem

    private class GeneticInformationItem {

      private Type myAncestorDefinition;
      public Type AncestorDefinition {
        get { return myAncestorDefinition; }
      }

      private int myAncestorIndex;
      public int AncestorIndex {
        get { return myAncestorIndex; }
      }

      private Type myDescendantDefinition;
      public Type DescendantDefinition {
        get { return myDescendantDefinition; }
      }

      private int myAncestorLevel;
      public int AncestorLevel {
        get { return myAncestorLevel; }
      }

      // not used in HL 3.3
      // private double myAncestorCoefficient;
      // public double AncestorCoefficient {
      //   get { return myAncestorCoefficient; }
      // }

      // not used in HL 3.3
      // private double myAncestorVariableIndex;
      // public double AncestorVariableIndex {
      //   get { return myAncestorVariableIndex; }
      // }

      // not used in HL 3.3
      // private int myAncestorTimeOffset;
      // public int AncestorTimeOffset {
      //   get { return myAncestorTimeOffset; }
      // }

      // not used in HL 3.3
      // private int myAncestorVariant;
      // public int AncestorVariant {
      //   get { return myAncestorVariant; }
      // }

      private double myDescendantCoefficient;
      public double DescendantCoefficient {
        get { return myDescendantCoefficient; }
      }

      private double myDescendantVariableIndex;
      public double DescendantVariableIndex {
        get { return myDescendantVariableIndex; }
      }

      private int myDescendantTimeOffset;
      public int DescendantTimeOffset {
        get { return myDescendantTimeOffset; }
      }

      private SymbolicExpressionTreeNode myDescendantTreeNode;
      public SymbolicExpressionTreeNode DescendantTreeNode {
        get { return myDescendantTreeNode; }
      }

      private int myDescendantLevel;
      public int DescendantLevel {
        get { return myDescendantLevel; }
      }

      public int LevelDelta {
        get { return (myDescendantLevel - myAncestorLevel); }
      }

      // not used in HL 3.3
      //private int myDescendantVariant;
      //public int DescendantVariant {
      //  get { return myDescendantVariant; }
      //}

      /*
      public static GeneticInformationItem FindBestPendant(GeneticInformationItem Item, List<GeneticInformationItem> ComparisonItems,
          StructuralSimilarityAnalysisParameters Parameters,
          int MaxTreeHeight, int MaxTimeOffset,
          out double BestPendantSimilarity) {
        int maxSimilarityIndex = -1;
        double similarity, maxSimilarity = -double.MaxValue;
        for (int i = 0; i < ComparisonItems.Count; i++) {
          similarity = Similarity(Item, ComparisonItems[i], Parameters, MaxTreeHeight, MaxTimeOffset);
          if (!double.IsNaN(similarity) && similarity > maxSimilarity) {
            maxSimilarity = similarity;
            maxSimilarityIndex = i;
          }
        }
        BestPendantSimilarity = maxSimilarity;
        if (maxSimilarityIndex >= 0)
          return ComparisonItems[maxSimilarityIndex];
        else
          return null;
      }*/

      /*
      public static double Similarity(GeneticInformationItem Item1, GeneticInformationItem Item2,
          StructuralSimilarityAnalysisParameters Parameters,
          int MaxTreeHeight, int MaxTimeOffset) {

        if (Item1.AncestorDefinition != Item2.AncestorDefinition ||
          Item1.DescendantDefinition != Item2.DescendantDefinition)
          return double.NaN;

        // the two items for sure have the same behavior definitions
        #region init
        double punishmentContributionSum = 0;
        double punishmentCoefficientsProduct = 1;
        double ancestorCoefficientDifferencePunishment = 0;
        double ancestorTimeOffsetDifferencePunishment = 0;
        double ancestorVariantDifferencePunishment = 0;
        double ancestorVariableIndexDifferencePunishment = 0;
        double descendantCoefficientDifferencePunishment = 0;
        double descendantTimeOffsetDifferencePunishment = 0;
        double descendantVariantDifferencePunishment = 0;
        double descendantVariableIndexDifferencePunishment = 0;
        double ancestorIndexDifferencePunishment = 0;
        double levelDifferencePunishment = 0;
        #endregion

        ITerminalDefinition ancestorTerminal = Item1.AncestorDefinition as ITerminalDefinition;
        ITerminalDefinition descendantTerminal = Item1.DescendantDefinition as ITerminalDefinition;
        IFunctionDefinition ancestorFunction = Item1.AncestorDefinition as IFunctionDefinition;
        IFunctionDefinition descendantFunction = Item1.DescendantDefinition as IFunctionDefinition;

        if (Parameters.ConsiderLevelDifference) {
          levelDifferencePunishment = Item1.LevelDelta - Item2.LevelDelta;
          if (levelDifferencePunishment < 0)
            levelDifferencePunishment = -levelDifferencePunishment;
          levelDifferencePunishment /= MaxTreeHeight;
          if (levelDifferencePunishment > 1)
            levelDifferencePunishment = 1;
          levelDifferencePunishment *= Parameters.LevelDifferenceCoefficient;
          punishmentContributionSum += Parameters.LevelDifferenceCoefficient;
          punishmentCoefficientsProduct *= (1 - Parameters.LevelDifferenceCoefficient);
        }
        if (Parameters.ConsiderAncestorIndex) {
          if (Item1.AncestorIndex != Item2.AncestorIndex)
            ancestorIndexDifferencePunishment = 1;
          else
            ancestorIndexDifferencePunishment = 0;
          ancestorIndexDifferencePunishment *= Parameters.AncestorIndexCoefficient;
          punishmentContributionSum += Parameters.AncestorIndexCoefficient;
          punishmentCoefficientsProduct *= (1 - Parameters.AncestorIndexCoefficient);
        }

        if (Item1.AncestorDefinition is ITerminalDefinition) {
          if (Parameters.ConsiderCoefficient) {
            double coefficientDifference = Math.Abs(Item1.myAncestorCoefficient - Item2.myAncestorCoefficient);
            if (ancestorTerminal.Parametrization.CoefficientIsGaussian)
              ancestorCoefficientDifferencePunishment = (coefficientDifference / (ancestorTerminal.Parametrization.CoefficientSigma * 4));
            else
              ancestorCoefficientDifferencePunishment = (coefficientDifference / (ancestorTerminal.Parametrization.CoefficientMaxValue - ancestorTerminal.Parametrization.CoefficientMinValue));
            if (ancestorCoefficientDifferencePunishment > 1)
              ancestorCoefficientDifferencePunishment = 1;
            ancestorCoefficientDifferencePunishment *= Parameters.CoefficientCoefficient;
            punishmentContributionSum += Parameters.CoefficientCoefficient;
            punishmentCoefficientsProduct *= (1 - Parameters.CoefficientCoefficient);
          }
          if (Parameters.ConsiderTimeOffset) {
            double timeOffsetDifference = Math.Abs(Item1.AncestorTimeOffset - Item2.AncestorTimeOffset);
            if (MaxTimeOffset > 0)
              ancestorTimeOffsetDifferencePunishment = timeOffsetDifference / MaxTimeOffset;
            ancestorTimeOffsetDifferencePunishment *= Parameters.TimeOffsetCoefficient;
            punishmentContributionSum += Parameters.TimeOffsetCoefficient;
            punishmentCoefficientsProduct *= (1 - Parameters.TimeOffsetCoefficient);
          }
          if (Parameters.ConsiderVariableIndex) {
            if (Item1.AncestorVariableIndex != Item2.AncestorVariableIndex)
              ancestorVariableIndexDifferencePunishment = 1;
            else
              ancestorVariableIndexDifferencePunishment = 0;
            ancestorVariableIndexDifferencePunishment *= Parameters.VariableIndexCoefficient;
            punishmentContributionSum += Parameters.VariableIndexCoefficient;
            punishmentCoefficientsProduct *= (1 - Parameters.VariableIndexCoefficient);
          }
        } else {
          if (Parameters.ConsiderVariant) {
            if (Item1.AncestorVariant != Item2.AncestorVariant)
              ancestorVariantDifferencePunishment = 1;
            else
              ancestorVariantDifferencePunishment = 0;
            ancestorVariantDifferencePunishment *= Parameters.VariantCoefficient;
            punishmentContributionSum += Parameters.VariantCoefficient;
            punishmentCoefficientsProduct *= (1 - Parameters.VariantCoefficient);
          }
        }

        if (Item1.DescendantDefinition is ITerminalDefinition) {
          if (Parameters.ConsiderCoefficient) {
            double coefficientDifference = Math.Abs(Item1.myDescendantCoefficient - Item2.myDescendantCoefficient);
            if (descendantTerminal.Parametrization.CoefficientIsGaussian)
              descendantCoefficientDifferencePunishment = (coefficientDifference / (descendantTerminal.Parametrization.CoefficientSigma * 4));
            else
              descendantCoefficientDifferencePunishment = (coefficientDifference / (descendantTerminal.Parametrization.CoefficientMaxValue - descendantTerminal.Parametrization.CoefficientMinValue));
            if (descendantCoefficientDifferencePunishment > 1)
              descendantCoefficientDifferencePunishment = 1;
            descendantCoefficientDifferencePunishment *= Parameters.CoefficientCoefficient;
            punishmentContributionSum += Parameters.CoefficientCoefficient;
            punishmentCoefficientsProduct *= (1 - Parameters.CoefficientCoefficient);
          }
          if (Parameters.ConsiderTimeOffset) {
            double timeOffsetDifference = Math.Abs(Item1.DescendantTimeOffset - Item2.DescendantTimeOffset);
            if (MaxTimeOffset > 0)
              descendantTimeOffsetDifferencePunishment = timeOffsetDifference / MaxTimeOffset;
            descendantTimeOffsetDifferencePunishment *= Parameters.TimeOffsetCoefficient;
            punishmentContributionSum += Parameters.TimeOffsetCoefficient;
            punishmentCoefficientsProduct *= (1 - Parameters.TimeOffsetCoefficient);
          }
          if (Parameters.ConsiderVariableIndex) {
            if (Item1.DescendantVariableIndex != Item2.DescendantVariableIndex)
              descendantVariableIndexDifferencePunishment = 1;
            else
              descendantVariableIndexDifferencePunishment = 0;
            descendantVariableIndexDifferencePunishment *= Parameters.VariableIndexCoefficient;
            punishmentContributionSum += Parameters.VariableIndexCoefficient;
            punishmentCoefficientsProduct *= (1 - Parameters.VariableIndexCoefficient);
          }
        } else {
          if (Parameters.ConsiderVariant) {
            if (Item1.DescendantVariant != Item2.DescendantVariant)
              descendantVariantDifferencePunishment = 1;
            else
              descendantVariantDifferencePunishment = 0;
            descendantVariantDifferencePunishment *= Parameters.VariantCoefficient;
            punishmentContributionSum += Parameters.VariantCoefficient;
            punishmentCoefficientsProduct *= (1 - Parameters.VariantCoefficient);
          }
        }

        double result;

        if (Parameters.AdditiveSimilarityCalculation) {
          double punishmentsSum =
            ancestorCoefficientDifferencePunishment + ancestorTimeOffsetDifferencePunishment +
            ancestorVariantDifferencePunishment + ancestorVariableIndexDifferencePunishment +
            descendantCoefficientDifferencePunishment + descendantTimeOffsetDifferencePunishment +
            descendantVariantDifferencePunishment + descendantVariableIndexDifferencePunishment +
            ancestorIndexDifferencePunishment + levelDifferencePunishment;
          result = (1 - punishmentsSum / punishmentContributionSum);
        } else {
          result =
            (1 - ancestorCoefficientDifferencePunishment) *
            (1 - ancestorTimeOffsetDifferencePunishment) *
            (1 - ancestorVariantDifferencePunishment) *
            (1 - ancestorVariableIndexDifferencePunishment) *
            (1 - descendantCoefficientDifferencePunishment) *
            (1 - descendantTimeOffsetDifferencePunishment) *
            (1 - descendantVariantDifferencePunishment) *
            (1 - descendantVariableIndexDifferencePunishment) *
            (1 - ancestorIndexDifferencePunishment) *
            (1 - levelDifferencePunishment);
          // worst possible result is (1-punishmentCoefficientsProduct), so scale linearly to [0;1]:
          result = (result - punishmentCoefficientsProduct) / (1 - punishmentCoefficientsProduct);
        }

        if (result < 0 || result > 1)
          throw new InvalidOperationException("ERROR in GeneticInformationItem.Similarity: An invalid similarity value (" + result.ToString() + ") has been calculated.");

        return result;

      }*/

      /*
      public static List<GeneticInformationItem> GetGeneticInformationItems(SymbolicExpressionTreeNode Formula, int MinLevelDifference, int MaxLevelDifference) {
        List<GeneticInformationItem> result = new List<GeneticInformationItem>();
        List<SymbolicExpressionTreeNode> allNodes = new List<SymbolicExpressionTreeNode>();
        allNodes.Add(Formula);
        allNodes.AddRange(getDescendants(Formula));
        foreach (SymbolicExpressionTreeNode node in allNodes) {
          List<SymbolicExpressionTreeNode> allDescendants = new List<SymbolicExpressionTreeNode>();
          allDescendants.Add(node);
          allDescendants.AddRange(getDescendants(node));
          foreach (SymbolicExpressionTreeNode descendant in allDescendants) {
            GeneticInformationItem item = new GeneticInformationItem(node, descendant);
            if (item.LevelDelta >= MinLevelDifference && item.LevelDelta <= MaxLevelDifference)
              result.Add(item);
          }
        }
        return result;
      }*/

      /*
      public static List<SymbolicExpressionTreeNode> GetDescendants(SymbolicExpressionTreeNode node) {
        List<SymbolicExpressionTreeNode> descendants = new List<SymbolicExpressionTreeNode>();
        foreach (SymbolicExpressionTreeNode subTreeNode in node.SubTrees) {
          AddDescendants(descendants, subTreeNode);
        }
        return descendants;
      }
      private static void AddDescendants(List<SymbolicExpressionTreeNode> list, SymbolicExpressionTreeNode node) {
        list.Add(node);
        foreach (SymbolicExpressionTreeNode subTreeNode in node.SubTrees) {
          AddDescendants(list, subTreeNode);
        }
      }*/

      public static IList<GeneticInformationItem> getGeneticInformationItems(SymbolicExpressionTreeNode node, List<string> variableNames,
          int MinimumLevelDelta, int MaximumLevelDelta) {
        // first we have to collect all items, then we filter; it is not possible to filter while collecting because the items are
        // collected recursively and used for collecting the parents' items.
        if (MinimumLevelDelta > MaximumLevelDelta)
          return new List<GeneticInformationItem>();
        IList<GeneticInformationItem> list = getGeneticInformationItems(node, variableNames, 0);
        List<GeneticInformationItem> resultList = new List<GeneticInformationItem>();
        foreach (GeneticInformationItem item in list)
          if (item.LevelDelta >= MinimumLevelDelta && item.LevelDelta <= MaximumLevelDelta)
            resultList.Add(item);
        return resultList;
      }

      private static IList<GeneticInformationItem> getGeneticInformationItems(SymbolicExpressionTreeNode node, List<string> variableNames, int level) {
        // Idea: collect all descendants' lists and then add new items using the retrieved ones.
        // This should save lots of time and reduce complexity of the items retrieval process.
        // Program roots are not considered, neither are start symbol nodes
        if (node.Symbol is ProgramRootSymbol)
          return getGeneticInformationItems(node.SubTrees[0], variableNames, level + 1);
        List<GeneticInformationItem> list = new List<GeneticInformationItem>();
        // add item for this node:
        if (!(node.Symbol is StartSymbol)) {
          list.Add(new GeneticInformationItem(node, variableNames, level));
        }
        // add items for the descendants, but prevent multiple references to descendant nodes:
        List<SymbolicExpressionTreeNode> descendantNodes = new List<SymbolicExpressionTreeNode>();
        for (int i = 0; i < node.SubTrees.Count; i++) {
          IList<GeneticInformationItem> descendantItems = getGeneticInformationItems(node.SubTrees[i], variableNames, level + 1);
          list.AddRange(descendantItems);
          if (!(node.Symbol is StartSymbol))
            foreach (GeneticInformationItem item in descendantItems) {
              if (!descendantNodes.Contains(item.DescendantTreeNode))
                list.Add(new GeneticInformationItem(node, item, i, level));
              descendantNodes.Add(item.DescendantTreeNode);
            }
        }
        return list;
      }

      private GeneticInformationItem (SymbolicExpressionTreeNode node, List<string> variableNames, int level) {
        myAncestorIndex = -1;
        VariableTreeNode variableTreeNode = node as VariableTreeNode;
        LaggedVariableTreeNode laggedVariableTreeNode = node as LaggedVariableTreeNode;
        ConstantTreeNode constantTreeNode = node as ConstantTreeNode;
        myAncestorDefinition = node.Symbol.GetType();
        myDescendantDefinition = myAncestorDefinition;
        if (variableTreeNode != null)
          myDescendantCoefficient = variableTreeNode.Weight;
        else if (constantTreeNode != null)
          myDescendantCoefficient = constantTreeNode.Value;
        else
          myDescendantCoefficient = double.NaN;
        if (laggedVariableTreeNode != null)
          myDescendantTimeOffset = laggedVariableTreeNode.Lag;
        else
          myDescendantTimeOffset = 0;
        if (variableTreeNode != null)
          myDescendantVariableIndex = variableNames.IndexOf(variableTreeNode.VariableName);
        else
          myDescendantVariableIndex = -1;
        myAncestorLevel = level;
        myDescendantLevel = level;
        myDescendantTreeNode = node;
      }

      private GeneticInformationItem(SymbolicExpressionTreeNode parentNode, GeneticInformationItem descendantGeneticInformationItem,
          int ancestorIndex, int parentNodeLevel) {
        myAncestorIndex = ancestorIndex;
        myAncestorLevel = parentNodeLevel;
        myAncestorDefinition = parentNode.Symbol.GetType();
        myDescendantCoefficient = descendantGeneticInformationItem.DescendantCoefficient;
        myDescendantDefinition = descendantGeneticInformationItem.DescendantDefinition;
        myDescendantTimeOffset = descendantGeneticInformationItem.DescendantTimeOffset;
        myDescendantVariableIndex = descendantGeneticInformationItem.DescendantVariableIndex;
        myDescendantLevel = descendantGeneticInformationItem.DescendantLevel;
        myDescendantTreeNode = descendantGeneticInformationItem.DescendantTreeNode;
      }

      public override string ToString() {
        return "ancestor: " + AncestorDefinition.Name.ToString() + ", [" + AncestorIndex + "]; "
          + "descendant: " + DescendantDefinition.Name.ToString() + " (varIndex " + DescendantVariableIndex + ", "
          + DescendantCoefficient + ", t-" + DescendantTimeOffset + ");"
          + " level delta = " + DescendantLevel + "-" + AncestorLevel + " = " + LevelDelta;
      }

    }

    #endregion

  }
}
